#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Modules for TransresnetMultimodalAgent.
"""

import torch
from torch import nn
from parlai.agents.transformer.modules import (
    TransformerEncoder,
    create_position_codes,
    TransformerEncoderLayer,
)
from projects.personality_captions.transresnet.modules import (
    TransresnetModel,
    load_fasttext_embeddings,
)


class TransresnetMultimodalModel(TransresnetModel):
    """
    Extension of Transresnet to incorporate dialogue history and multimodality.
    """

    @staticmethod
    def add_cmdline_args(argparser):
        """
        Override to include model-specific args.
        """
        TransresnetModel.add_cmdline_args(argparser)
        agent = argparser.add_argument_group("TransresnetMultimodal task arguments")
        agent.add_argument(
            "--context-encoder-embedding-type",
            type=str,
            default=None,
            choices=[None, "fasttext_cc"],
            help="Specify if using pretrained embeddings",
        )
        agent.add_argument(
            "--load-context-encoder-from",
            type=str,
            default=None,
            help="Specify if using a pretrained transformer encoder",
        )
        agent.add_argument(
            "--share-encoder",
            type="bool",
            default=False,
            help="Whether to share the text encoder for the "
            "labels and the dialogue history",
        )
        agent.add_argument("--num-layers-multimodal-encoder", type=int, default=1)
        agent.add_argument(
            "--multimodal",
            type="bool",
            default=False,
            help="If true, feed a query term into a separate "
            "transformer prior to computing final rank "
            "scores",
        )
        agent.add_argument(
            "--multimodal-combo",
            type=str,
            choices=["concat", "sum"],
            default="sum",
            help="How to combine the encoding for the " "multi-modal transformer",
        )
        agent.add_argument(
            "--encode-image",
            type="bool",
            default=True,
            help="Whether to include the image encoding when "
            "retrieving a candidate response",
        )
        agent.add_argument(
            "--encode-dialogue-history",
            type="bool",
            default=True,
            help="Whether to include the dialogue history "
            "encoding when retrieving a candidate response",
        )
        agent.add_argument(
            "--encode-personality",
            type="bool",
            default=True,
            help="Whether to include the personality encoding "
            "when retrieving a candidate response",
        )

    def __init__(self, opt, personalities_list, dictionary):
        super().__init__(opt, personalities_list, dictionary)
        self.hidden_dim = self.opt["hidden_dim"]
        self.share_encoder = opt.get("share_encoder")
        nlayers_mm = (
            opt["num_layers_all"]
            if opt["num_layers_all"] != -1
            else opt["num_layers_multimodal_encoder"]
        )

        # blank encoding (for concat)
        self.blank_encoding = torch.Tensor(opt["hidden_dim"]).fill_(0).detach_()
        if self.use_cuda:
            self.blank_encoding = self.blank_encoding.cuda()

        # Encoders
        self.encode_image = opt.get("encode_image", True)
        self.encode_personality = opt.get("encode_personality", True)
        self.encode_dialogue_history = opt.get("encode_dialogue_history", True)
        assert any(
            [self.encode_dialogue_history, self.encode_image, self.encode_personality]
        )

        # Transformer 2
        self._build_multimodal_encoder(nlayers_mm)

        # Label Encoder
        self.label_encoder = self.text_encoder

        # Context encoder
        self._build_context_encoder()

    def _build_multimodal_encoder(self, n_layers_mm):
        """
        Build the multimodal encoder.

        :param n_layers_mm:
            number of layers for the transformer
        """
        self.multimodal = self.opt.get("multimodal")
        if self.multimodal:
            self.multimodal_combo = self.opt.get("multimodal_combo", "sum")
            nlayers_mm = (
                self.opt["num_layers_all"]
                if self.opt["num_layers_all"] != -1
                else self.opt["num_layers_multimodal_encoder"]
            )
            self.multimodal_encoder = MultimodalCombiner(
                n_heads=self.opt["n_heads"],
                n_layers=nlayers_mm,
                hidden_dim=self.opt["hidden_dim"],
                ffn_size=self.opt["embedding_size"] * 4,
                attention_dropout=self.opt["attention_dropout"],
                relu_dropout=self.opt["relu_dropout"],
                learn_positional_embeddings=self.opt.get(
                    "learn_positional_embeddings", False
                ),
                reduction=True,
            )

    def _build_context_encoder(self):
        """
        Build the context (i.e. dialogue history) encoder.
        """
        if self.opt.get("share_encoder"):
            self.context_encoder = self.label_encoder
        else:
            if (
                self.opt["load_context_encoder_from"] is None
                and self.opt["context_encoder_embedding_type"] == "fasttext_cc"
            ):
                embeddings = load_fasttext_embeddings(
                    self.dictionary, self.opt["embedding_size"], self.opt["datapath"]
                )
            else:
                embeddings = nn.Embedding(
                    len(self.dictionary), self.opt["embedding_size"]
                )
            self.context_encoder = TransformerEncoder(
                n_heads=self.opt["n_heads"],
                n_layers=self.opt["n_layers"],
                embedding_size=self.opt["embedding_size"],
                ffn_size=self.opt["ffn_size"],
                vocabulary_size=len(self.dictionary),
                embedding=embeddings,
                dropout=self.opt["dropout"],
                attention_dropout=self.opt["attention_dropout"],
                relu_dropout=self.opt["relu_dropout"],
                padding_idx=self.dictionary.tok2ind[self.dictionary.null_token],
                learn_positional_embeddings=self.opt["learn_positional_embeddings"],
                embeddings_scale=False,
                n_positions=self.opt["n_positions"],
                activation=self.opt["activation"],
                variant=self.opt["variant"],
                n_segments=self.opt["n_segments"],
            )
            if self.opt.get("load_context_encoder_from") is not None:
                self._load_context_encoder_state()

    def forward(
        self,
        image_features,
        personalities,
        dialogue_histories,
        labels,
        batchsize=None,
        personalities_tensor=None,
    ):
        """
        Model forward pass.

        :param image_features:
            list of tensors of image features, one per example
        :param personalities:
            list of personalities, one per example
        :param dialogue_histories:
            list of dialogue histories, one per example
        :param labels:
            list of response labels, one per example
        :param personalities_tensor:
            (optional) list of personality representations, usually a one-hot
            vector if specified

        :return:
            the encoded context and the encoded captions.
        """
        # labels
        labels_encoded = self.forward_text_encoder(labels)
        # dialog history
        d_hist_encoded = self.forward_text_encoder(
            dialogue_histories, dialogue_history=True, batchsize=batchsize
        )
        # images
        img_encoded = self.forward_image(image_features)
        # personalities
        pers_encoded = self.forward_personality(personalities, personalities_tensor)
        total_encoded = self.get_rep(
            [img_encoded, d_hist_encoded, pers_encoded], batchsize=batchsize
        )
        loss, nb_ok = self.get_loss(total_encoded, labels_encoded)

        return loss, nb_ok, total_encoded

    def forward_personality(self, personalities, personalities_tensor):
        """
        Encode personalities.

        :param personalities:
            list of personalities, one per example
        :param personalities_tensor:
            (optional) list of personality representations, usually a one-hot
            vector if specified

        :return:
            encoded representation of the personalities
        """
        pers_encoded = None
        if not self.encode_personality:
            if self.multimodal and self.multimodal_combo == "concat":
                pers_encoded = self.blank_encoding
        else:
            pers_encoded = super().forward_personality(
                personalities, personalities_tensor
            )
        return pers_encoded

    def forward_text_encoder(self, texts, dialogue_history=False, batchsize=None):
        """
        Forward pass for a text encoder.

        :param texts:
            text to encode
        :param dialogue_history:
            flag that indicates whether the text is dialogue history; if False,
            text is a response candidate
        :param batchsize:
            size of the batch

        :return:
            encoded representation of the `texts`
        """
        texts_encoded = None
        if texts is None or (dialogue_history and not self.encode_dialogue_history):
            if (
                self.multimodal
                and self.multimodal_combo == "concat"
                and dialogue_history
            ):
                texts_encoded = torch.stack(
                    [self.blank_encoding for _ in range(batchsize)]
                )
        else:
            encoder = self.context_encoder if dialogue_history else self.label_encoder
            indexes, mask = self.captions_to_tensor(texts)
            texts_encoded = encoder(indexes)
            if self.text_encoder_frozen:
                texts_encoded = texts_encoded.detach()
            texts_encoded = self.additional_layer(texts_encoded)

        return texts_encoded

    def forward_image(self, image_features):
        """
        Encode image features.

        :param image_features:
            list of image features

        :return:
            encoded representation of the image features
        """
        img_encoded = None
        if image_features is None or not self.encode_image:
            if self.multimodal and self.multimodal_combo == "concat":
                img_encoded = self.blank_encoding
        else:
            img_encoded = super().forward_image(image_features)

        return img_encoded

    def get_rep(self, encodings, batchsize=None):
        """
        Get the multimodal representation of the encodings.

        :param encodings:
            list of encodings
        :param batchsize:
            size of batch

        :return:
            final multimodal representations
        """
        if not self.multimodal:
            rep = self.sum_encodings(encodings)
        else:
            if self.multimodal_combo == "sum":
                encodings = self.sum_encodings(encodings).unsqueeze(1)
            elif self.multimodal_combo == "concat":
                encodings = self.cat_encodings(encodings)
            all_one_mask = torch.ones(encodings.size()[:2])
            if self.use_cuda:
                all_one_mask = all_one_mask.cuda()
            rep = self.multimodal_encoder(encodings, all_one_mask)
        if rep is None:
            rep = torch.stack([self.blank_encoding for _ in range(batchsize)])
        return rep

    def choose_best_response(
        self,
        image_features,
        personalities,
        dialogue_histories,
        candidates,
        candidates_encoded=None,
        k=1,
        batchsize=None,
    ):
        """
        Choose the best response for each example.

        :param image_features:
            list of tensors of image features
        :param personalities:
            list of personalities
        :param dialogue_histories:
            list of dialogue histories, one per example
        :param candidates:
            list of candidates, one set per example
        :param candidates_encoded:
            optional; if specified, a fixed set of encoded candidates that is
            used for each example
        :param k:
            number of ranked candidates to return. if < 1, we return the ranks
            of all candidates in the set.

        :return:
            a set of ranked candidates for each example
        """
        self.eval()
        _, _, encoded = self.forward(
            image_features, personalities, dialogue_histories, None, batchsize=batchsize
        )
        encoded = encoded.detach()
        one_cand_set = True
        if candidates_encoded is None:
            one_cand_set = False
            candidates_encoded = [
                self.forward_text_encoder(c).detach() for c in candidates
            ]
        chosen = [
            self.choose_topk(
                idx if not one_cand_set else 0,
                encoded,
                candidates,
                candidates_encoded,
                one_cand_set,
                k,
            )
            for idx in range(len(encoded))
        ]
        return chosen

    def choose_topk(
        self, idx, encoded, candidates, candidates_encoded, one_cand_set, k
    ):
        """
        Choose top k best responses for a single example.

        :param idx:
            idx of example in encoded
        :param encoded:
            full matrix of encoded representations (for the whole batch)
        :param candidates:
            list of candidates
        :param candidates_encoded:
            encoding of the candidates
        :param one_cand_set:
            true if there is one set of candidates for each example
        :param k:
            how many ranked responses to return

        :return:
            ranked list of k responses
        """
        encoding = encoded[idx : idx + 1, :]
        scores = torch.mm(
            candidates_encoded[idx] if not one_cand_set else candidates_encoded,
            encoding.transpose(0, 1),
        )
        if k >= 1:
            _, index_top = torch.topk(scores, k, dim=0)
        else:
            _, index_top = torch.topk(scores, scores.size(0), dim=0)
        return [
            candidates[idx][idx2] if not one_cand_set else candidates[idx2]
            for idx2 in index_top.unsqueeze(1)
        ]

    def get_loss(self, total_encoded, labels_encoded):
        """
        Compute loss over batch.

        :param total_encoded:
            encoding of the examples
        :param labels_encoded:
            encoding of the labels

        :return:
            total batch loss, and number of correct examples
        """
        loss = None
        num_correct = None
        if labels_encoded is not None:
            dot_products = total_encoded.mm(
                labels_encoded.t()
            )  # batch_size * batch_size
            log_prob = torch.nn.functional.log_softmax(dot_products, dim=1)
            targets = torch.arange(0, len(total_encoded), dtype=torch.long)
            if self.use_cuda:
                targets = targets.cuda()
            loss = torch.nn.functional.nll_loss(log_prob, targets)
            num_correct = (log_prob.max(dim=1)[1] == targets).float().sum()
        return loss, num_correct

    def cat_encodings(self, tensors):
        """
        Concatenate non-`None` encodings.

        :param tensors:
            list tensors to concatenate

        :return:
            concatenated tensors
        """
        tensors = [t for t in tensors if t is not None]
        return torch.cat([t.unsqueeze(1) for t in tensors], dim=1)

    def _load_text_encoder_state(self):
        try:
            state_file = self.opt.get("load_encoder_from")
            model = torch.load(state_file)
            states = model["model"]
            self.text_encoder.load_state_dict(states)
        except Exception as e:
            print(
                "WARNING: Cannot load transformer state; please make sure "
                "specified file is a dictionary with the states in `model`. "
                "Additionally, make sure that the appropriate options are "
                "specified. Error: {}".format(e)
            )

    def _load_context_encoder_state(self):
        try:
            state_file = self.opt.get("load_context_encoder_from")
            model = torch.load(state_file)
            states = model["model"]
            self.context_encoder.load_state_dict(states)
        except Exception as e:
            print(
                "WARNING: Cannot load transformer state; please make sure "
                "specified file is a dictionary with the states in `model`. "
                "Additionally, make sure that the appropriate options are "
                "specified. Error: {}".format(e)
            )


class MultimodalCombiner(nn.Module):
    """
    Multimodal Combination module.
    """

    def __init__(
        self,
        n_heads,
        n_layers,
        hidden_dim,
        ffn_size,
        reduction=True,
        attention_dropout=0.0,
        relu_dropout=0.0,
        learn_positional_embeddings=False,
    ):
        super().__init__()
        self.ffn_size = ffn_size
        self.n_layers = n_layers
        self.n_heads = n_heads
        self.out_dim = hidden_dim
        self.dim = hidden_dim
        self.reduction = reduction
        assert hidden_dim % n_heads == 0, "MM-Combiner dim must be multiple of n_heads"
        n_positions = 1024
        self.position_embeddings = nn.Embedding(n_positions, hidden_dim)
        if not learn_positional_embeddings:
            create_position_codes(
                n_positions, hidden_dim, out=self.position_embeddings.weight
            )
        else:
            nn.init.normal_(self.position_embeddings.weight, 0, hidden_dim ** -0.5)

        self.layers = nn.ModuleList()
        for _ in range(self.n_layers):
            self.layers.append(
                TransformerEncoderLayer(
                    n_heads, hidden_dim, ffn_size, attention_dropout, relu_dropout
                )
            )

    def forward(self, tensor, mask):
        """
        Forward pass.

        :param tensor:
            a [bsz, seq_len, hidden_dim] FloatTensor
        :param mask:
            a [bsz, seq_len] ByteTensor filled with 1 when inside the sequence and 0 outside.

        :return:
            output: a [bsz, hidden_dim] FloatTensor of encodings
            mask: the same as before
        """
        seq_len = tensor.size(1)
        positions = tensor.new(seq_len).long()
        positions = torch.arange(seq_len, out=positions).unsqueeze(0)
        tensor = tensor + self.position_embeddings(positions).expand_as(tensor)

        tensor *= mask.unsqueeze(-1).float()
        for i in range(self.n_layers):
            tensor = self.layers[i](tensor, mask)

        if self.reduction:
            divisor = mask.float().sum(dim=1).unsqueeze(-1).clamp(min=1e-20)
            output = tensor.sum(dim=1) / divisor
            return output
        else:
            output = tensor
            return output, mask



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Provide functionality for loading images.
"""

import parlai.core.build_data as build_data
import parlai.utils.logging as logging
from parlai.utils.io import PathManager

import os
from PIL import Image
from zipfile import ZipFile

_greyscale = '  .,:;crsA23hHG#98&@'
_cache_size = 84000

# Mapping from image mode to (torch_instantiation_str, layer_cutoff_idx)
IMAGE_MODE_SWITCHER = {
    'resnet152': ['resnet152', -1],
    'resnet101': ['resnet101', -1],
    'resnet50': ['resnet50', -1],
    'resnet34': ['resnet34', -1],
    'resnet18': ['resnet18', -1],
    'resnet152_spatial': ['resnet152', -2],
    'resnet101_spatial': ['resnet101', -2],
    'resnet50_spatial': ['resnet50', -2],
    'resnet34_spatial': ['resnet34', -2],
    'resnet18_spatial': ['resnet18', -2],
    'resnext101_32x8d_wsl': ['resnext101_32x8d_wsl', -1],
    'resnext101_32x16d_wsl': ['resnext101_32x16d_wsl', -1],
    'resnext101_32x32d_wsl': ['resnext101_32x32d_wsl', -1],
    'resnext101_32x48d_wsl': ['resnext101_32x48d_wsl', -1],
    'resnext101_32x8d_wsl_spatial': ['resnext101_32x8d_wsl', -2],
    'resnext101_32x16d_wsl_spatial': ['resnext101_32x16d_wsl', -2],
    'resnext101_32x32d_wsl_spatial': ['resnext101_32x32d_wsl', -2],
    'resnext101_32x48d_wsl_spatial': ['resnext101_32x48d_wsl', -2],
}


class ImageLoader:
    """
    Extract image feature using pretrained CNN network.
    """

    def __init__(self, opt):
        self.opt = opt.copy()
        self.use_cuda = False
        self.netCNN = None
        self.im = opt.get('image_mode', 'no_image_model')
        if self.im not in ['no_image_model', 'raw', 'ascii']:
            if 'image_mode' not in opt or 'image_size' not in opt:
                raise RuntimeError(
                    'Need to add image arguments to opt. See '
                    'parlai.core.params.ParlaiParser.add_image_args'
                )
            self.image_mode = opt['image_mode']
            self.image_size = opt['image_size']
            self.crop_size = opt['image_cropsize']
            self._lazy_import_torch()
            self._init_transform()
            if 'resnet' in self.image_mode:
                self._init_resnet_cnn()
            elif 'resnext' in self.image_mode:
                self._init_resnext_cnn()
            else:
                raise RuntimeError(
                    'Image mode {} not supported'.format(self.image_mode)
                )

    def _lazy_import_torch(self):
        try:
            import torch
        except ImportError:
            raise ImportError('Need to install Pytorch: go to pytorch.org')
        import torchvision
        import torchvision.transforms as transforms
        import torch.nn as nn

        self.use_cuda = not self.opt.get('no_cuda', False) and torch.cuda.is_available()
        if self.use_cuda:
            logging.debug('Using CUDA')
            torch.cuda.set_device(self.opt.get('gpu', -1))
        self.torch = torch
        self.torchvision = torchvision
        self.transforms = transforms
        self.nn = nn

    def _init_transform(self):
        # initialize the transform function using torch vision.
        self.transform = self.transforms.Compose(
            [
                self.transforms.Scale(self.image_size),
                self.transforms.CenterCrop(self.crop_size),
                self.transforms.ToTensor(),
                self.transforms.Normalize(
                    mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                ),
            ]
        )

    def _init_resnet_cnn(self):
        """
        Lazily initialize preprocessor model.

        When image_mode is one of the ``resnet`` varieties
        """
        cnn_type, layer_num = self._image_mode_switcher()
        # initialize the pretrained CNN using pytorch.
        CNN = getattr(self.torchvision.models, cnn_type)

        # cut off the additional layer.
        self.netCNN = self.nn.Sequential(
            *list(CNN(pretrained=True).children())[:layer_num]
        )

        if self.use_cuda:
            self.netCNN.cuda()

    def _init_resnext_cnn(self):
        """
        Lazily initialize preprocessor model.

        When image_mode is one of the ``resnext101_..._wsl`` varieties
        """
        try:
            cnn_type, layer_num = self._image_mode_switcher()
            model = self.torch.hub.load('facebookresearch/WSL-Images', cnn_type)
            # cut off layer for ImageNet classification
            # if spatial, cut off another layer for spatial features
            self.netCNN = self.nn.Sequential(*list(model.children())[:layer_num])
        except RuntimeError as e:
            # Perhaps specified one of the wrong model names
            model_names = [m for m in IMAGE_MODE_SWITCHER if 'resnext101' in m]
            logging.error(
                'If you have specified one of the resnext101 wsl models, '
                'please make sure it is one of the following: \n'
                f"{', '.join(model_names)}"
            )
            raise e
        except AttributeError:
            # E.g. "module 'torch' has no attribute 'hub'"
            raise RuntimeError(
                'Please install the latest pytorch distribution to have access '
                'to the resnext101 wsl models (pytorch 1.1.0, torchvision 0.3.0)'
            )

        if self.use_cuda:
            self.netCNN.cuda()

    def _image_mode_switcher(self):
        if self.image_mode not in IMAGE_MODE_SWITCHER:
            raise NotImplementedError(
                'image preprocessing mode'
                + '{} not supported yet'.format(self.image_mode)
            )

        return IMAGE_MODE_SWITCHER.get(self.image_mode)

    @classmethod
    def get_available_model_names(cls):
        """
        Get a list of the available model variants in this ImageLoader.
        """
        return list(IMAGE_MODE_SWITCHER.keys())

    def extract(self, image, path=None):
        # check whether initialize CNN network.
        if not self.netCNN:
            self.init_cnn(self.opt)
        # extract the image feature
        transform = self.transform(image).unsqueeze(0)
        if self.use_cuda:
            transform = transform.cuda()
        with self.torch.no_grad():
            feature = self.netCNN(transform)
        # save the feature
        if path is not None:
            self.torch.save(feature.cpu(), path)
        return feature

    def _img_to_ascii(self, im):
        im.thumbnail((60, 40), Image.BICUBIC)
        im = im.convert('L')
        asc = []
        for y in range(0, im.size[1]):
            for x in range(0, im.size[0]):
                lum = 255 - im.getpixel((x, y))
                asc.append(_greyscale[lum * len(_greyscale) // 256])
            asc.append('\n')
        return ''.join(asc)

    def _breakup_zip_filename(self, path):
        # assume format path/to/file.zip/image_name.jpg
        assert '.zip' in path
        sep = path.index('.zip') + 4
        zipname = path[:sep]
        file_name = path[sep + 1 :]
        return zipname, file_name

    def _get_prepath(self, path):
        if '.zip' in path:
            zipname, file_name = self._breakup_zip_filename(path)
            task = self.opt['task']
            prepath = os.path.join(self.opt['datapath'], task)
            imagefn = ''.join(zipname.strip('.zip').split('/')[-2:]) + path.name
            return prepath, imagefn
        else:
            prepath, imagefn = os.path.split(path)
            return prepath, imagefn

    def _load_image(self, path):
        """
        Return the loaded image in the path.
        """
        if '.zip' in path:
            zipname, file_name = self._breakup_zip_filename(path)
            with ZipFile(PathManager.open(zipname, 'rb')) as zipf:
                with zipf.open(file_name) as fh:
                    return Image.open(fh).convert('RGB')
        else:
            # raw just returns RGB values
            with PathManager.open(path, 'rb') as f:
                return Image.open(f).convert('RGB')

    def load(self, path):
        """
        Load from a given path.
        """
        mode = self.opt.get('image_mode', 'raw')
        if mode is None or mode == 'no_image_model':
            # don't need to load images
            return None
        elif mode == 'raw':
            return self._load_image(path)
        elif mode == 'ascii':
            # convert images to ascii ¯\_(ツ)_/¯
            return self._img_to_ascii(self._load_image(path))

        # otherwise, looks for preprocessed version under 'mode' directory
        prepath, imagefn = self._get_prepath(path)
        dpath = os.path.join(prepath, mode)
        if not PathManager.exists(dpath):
            build_data.make_dir(dpath)
        imagefn = imagefn.split('.')[0]
        new_path = os.path.join(prepath, mode, imagefn)
        if not PathManager.exists(new_path):
            return self.extract(self._load_image(path), new_path)
        else:
            return self.torch.load(new_path)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.mturk.core.worlds import MTurkOnboardWorld
from parlai.core.worlds import validate
from joblib import Parallel, delayed
import numpy as np
import time
import os
import pickle


class PersonaProfileWorld(MTurkOnboardWorld):
    """
    A world that provides a persona to the MTurkAgent.
    """

    def __init__(self, opt, mturk_agent):
        self.opt = opt
        self.task_type = 'sandbox' if opt['is_sandbox'] else 'live'
        self.max_persona_time = opt['max_persona_time']
        self.range_persona = [int(s) for s in opt['range_persona'].split(',')]
        self.n_persona = np.random.randint(
            self.range_persona[0], self.range_persona[1] + 1
        )
        self.episodeDone = False
        self.persona = []
        self.persone_done = False
        super().__init__(opt, mturk_agent)

    def parley(self):
        # Persona creation instructions
        self.mturk_agent.observe(
            {
                'id': 'SYSTEM',
                'text': (
                    'Please create your character by entering'
                    '<b><span style="color:blue">{} sentences</span></b> below in'
                    'the input-box. \nYou have <b><span style="color:blue">{} '
                    'mins</span></b> to finish the persona creation.'.format(
                        self.n_persona, int(self.max_persona_time / 60)
                    )
                ),
            }
        )
        while not self.persona_done:
            act = self.mturk_agent.act(timeout=self.max_persona_time)
            # Check timeout
            if act['episode_done']:
                self.episodeDone = True
                return

            candidate_persona = [
                x for x in act['text'].split('.'.strip()) if x != ''  # sic.
            ]

            for cand in candidate_persona:
                # Check if persona is too long
                if len(cand.split(' ')) > 16:
                    control_msg = {
                        'id': 'SYSTEM',
                        'text': (
                            '\n A sentence you entered is too long:\n \
                            <b><span style="color:blue">'
                            + cand
                            + '</span></b>\nPlease resend a sentence '
                            '<b><span style="color:blue">less than 15 '
                            'words</span></b>.'
                        ),
                    }
                    self.mturk_agent.observe(validate(control_msg))
                    candidate_persona.remove(cand)
                # Check if persona is too short
                if len(cand.split(' ')) < 3:
                    control_msg = {
                        'id': 'SYSTEM',
                        'text': (
                            '\n A sentence you entered is too short:\n'
                            '<b><span style="color:blue">'
                            + cand
                            + '</span></b>\n Please resend a sentence '
                            '<b><span style="color:blue">more than 3 '
                            'words</span></b>.'
                        ),
                    }
                    self.mturk_agent.observe(validate(control_msg))
                    candidate_persona.remove(cand)

            self.persona += candidate_persona

            if len(self.persona) >= self.n_persona:
                self.persona_done = True
                control_msg = {
                    'id': 'SYSTEM',
                    'text': (
                        'Thank you for creating the character! \n '
                        '<b><span style="color:blue">Please click '
                        '"Done with this HIT" below to submit the HIT.'
                        '</span></b>'
                    ),
                    'exceed_min_turns': True,
                }
                self.mturk_agent.observe(validate(control_msg))
                self.episodeDone = True

            if not self.persona_done:
                control_msg = {
                    'id': 'SYSTEM',
                    'text': (
                        'Please enter at least *{}* more sentence(s) to finish. '.format(
                            str(self.n_persona - len(self.persona))
                        )
                    ),
                }
                self.mturk_agent.observe(validate(control_msg))

    def save_data(self):
        data_path = self.opt['data_path'] + '/personas'
        if not os.path.exists(data_path):
            os.makedirs(data_path)
        filename = os.path.join(
            data_path,
            'persona_{}_{}_{}.pkl'.format(
                time.strftime("%Y%m%d-%H%M%S"),
                self.mturk_agent.worker_id,
                self.task_type,
            ),
        )
        print('Profile successfully saved at {}.'.format(filename))
        pickle.dump(
            {
                'hit_id': self.mturk_agent.hit_id,
                'assignment_id': self.mturk_agent.assignment_id,
                'worker_id': self.mturk_agent.worker_id,
                'n_persona': self.n_persona,
                'persona': self.persona,
            },
            open(filename, 'wb'),
        )

    def shutdown(self):
        global shutdown_agent

        def shutdown_agent(mturk_agent):
            mturk_agent.shutdown()

        Parallel(n_jobs=1, backend='threading')(
            delayed(shutdown_agent)(agent) for agent in [self.mturk_agent]
        )

    def review_work(self):
        if self.persona_done:
            self.mturk_agent.approve_work()

    def episode_done(self):
        return self.episodeDone



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging

from parlai.mturk.core.agents import MTurkAgent, TIMEOUT_MESSAGE
from parlai.mturk.core.shared_utils import AssignState
import parlai.mturk.core.shared_utils as shared_utils


class MockTurkAgent(MTurkAgent):
    """
    Mock turk agent that can act in a parlai mturk world.
    """

    def __init__(self, opt, mturk_manager, hit_id, assignment_id, worker_id):
        super().__init__(opt, mturk_manager, hit_id, assignment_id, worker_id)
        self.db_logger = None
        self.mock_status = AssignState.STATUS_NONE
        self.wants_message = False
        self.unread_messages = []
        self.timed_out = False

    def get_update_packet(self):
        """
        Produce an update packet that represents the state change of this agent.
        """
        send_messages = []
        while len(self.unread_messages) > 0:
            pkt = self.unread_messages.pop(0)
            send_messages.append(pkt.data)
        done_text = None
        if self.state.is_final() and self.get_status() != AssignState.STATUS_DONE:
            done_text = self.state.get_inactive_command_text()[0]
        return {
            'new_messages': send_messages,
            'all_messages': self.state.get_messages(),
            'wants_message': self.wants_message,
            'disconnected': self.disconnected,
            'agent_id': self.id,
            'worker_id': self.worker_id,
            'conversation_id': self.conversation_id,
            'task_done': self.state.is_final(),
            'done_text': done_text,
            'status': self.state.get_status(),
        }

    def log_reconnect(self):
        """
        We aren't logging behavior in the mock.
        """
        pass

    def is_in_task(self):
        return self.status == AssignState.STATUS_IN_TASK

    def put_data(self, id, data):
        """
        Put data into the message queue.
        """
        self.msg_queue.put(data)

    def flush_msg_queue(self):
        """
        Clear all messages in the message queue.
        """
        if self.msg_queue is None:
            return
        while not self.msg_queue.empty():
            self.msg_queue.get()

    def prepare_timeout(self):
        """
        Log a timeout event, tell mturk manager it occurred, return message to return
        for the act call.
        """
        shared_utils.print_and_log(
            logging.INFO, '{} timed out before sending.'.format(self.id)
        )
        self.timed_out = True
        return self._get_episode_done_msg(TIMEOUT_MESSAGE)

    def request_message(self):
        if not (
            self.disconnected or self.some_agent_disconnected or self.hit_is_expired
        ):
            self.wants_message = True

    def act(self, timeout=None, blocking=True):
        """
        Retrieve an act in the normal expected way (out of the queue), but clear the act
        request if we do end up getting an act.
        """
        gotten_act = super().act(timeout, blocking)
        if gotten_act is not None:
            self.wants_message = False
            gotten_act['episode_done'] = gotten_act.get('episode_done', False)
        return gotten_act

    def episode_done(self):
        """
        Return whether or not this agent believes the conversation to be done.
        """
        if self.get_status() == AssignState.STATUS_DONE:
            return False
        else:
            return True

    def approve_work(self):
        print('[mock] Worker {} approved'.format(self.worker_id))

    def reject_work(self, reason='unspecified'):
        print('[mock] Worker {} rejected for reason {}'.format(self.worker_id, reason))

    def block_worker(self, reason='unspecified'):
        print('[mock] Worker {} blocked for reason {}'.format(self.worker_id, reason))

    def pay_bonus(self, bonus_amount, reason='unspecified'):
        print(
            '[mock] Worker {} bonused {} for reason {}'.format(
                self.worker_id, bonus_amount, reason
            )
        )

    def email_worker(self, subject, message_text):
        return True

    def set_hit_is_abandoned(self):
        self.hit_is_abandoned = True

    def wait_for_hit_completion(self, timeout=None):
        pass

    def shutdown(self, timeout=None, direct_submit=False):
        pass

    def update_agent_id(self, agent_id):
        """
        State is sent directly from the agent, so no need to send like MTurkAgent does
        in the full version.
        """
        self.id = agent_id



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.core.params import ParlaiParser
from parlai.mturk.tasks.react_task_demo.react_custom_no_extra_deps.worlds import (
    AskerOnboardingWorld,
    AnswererOnboardingWorld,
    EvaluatorOnboardingWorld,
    MultiRoleAgentWorld,
)
from parlai.mturk.core.mturk_manager import MTurkManager
from parlai.mturk.tasks.react_task_demo.react_custom_no_extra_deps.task_config import (
    task_config,
)
import os


def main():
    """
    Handles setting up and running a ParlAI-MTurk task by instantiating an MTurk manager
    and configuring it for the qa_data_collection task.
    """
    # Get relevant arguments
    argparser = ParlaiParser(False, False)
    argparser.add_parlai_data_path()
    argparser.add_mturk_args()
    opt = argparser.parse_args()

    # Set the task name to be the folder name
    opt['task'] = os.path.basename(os.path.dirname(os.path.abspath(__file__)))

    # append the contents of task_config.py to the configuration
    opt.update(task_config)

    # Select an agent_id that worker agents will be assigned in their world
    mturk_agent_roles = ['Asker', 'Answerer', 'Evaluator']

    # Instantiate an MTurkManager with the given options and a maximum number
    # of agents per world of 1 (based on the length of mturk_agent_ids)
    mturk_manager = MTurkManager(
        opt=opt, mturk_agent_ids=mturk_agent_roles, use_db=True
    )

    mturk_manager.setup_server(
        task_directory_path=os.path.dirname(os.path.abspath(__file__))
    )

    role_index = 0

    # Create an onboard_function, which will be run for workers who have
    # accepted your task and must be completed before they are put in the
    # queue for a task world.
    def run_onboard(worker):
        nonlocal role_index
        role = mturk_agent_roles[role_index % 3]
        role_index += 1
        worker.update_agent_id('Onboarding {}'.format(role))
        worker.demo_role = role
        if role == 'Asker':
            world = AskerOnboardingWorld(opt=opt, mturk_agent=worker)
        elif role == 'Answerer':
            world = AnswererOnboardingWorld(opt=opt, mturk_agent=worker)
        else:
            world = EvaluatorOnboardingWorld(opt=opt, mturk_agent=worker)
        while not world.episode_done():
            world.parley()
        world.shutdown()
        return world.prep_save_data([worker])

    # If we want to use the above onboard function, we can replace the below
    # with set_onboard_function(onboard_function=run_onboard)
    mturk_manager.set_onboard_function(onboard_function=run_onboard)

    try:
        # Initialize run information
        mturk_manager.start_new_run()

        # Set up the sockets and threads to recieve workers
        mturk_manager.ready_to_accept_workers()

        # Create the hits as specified by command line arguments
        mturk_manager.create_hits()

        # Check workers eligiblity acts as a filter, and should return
        # the list of all workers currently eligible to work on the task
        # Can be used to pair workers that meet certain criterea
        def check_workers_eligibility(workers):
            filled_roles = []
            use_workers = []
            for worker in workers:
                if worker.demo_role not in filled_roles:
                    use_workers.append(worker)
                    filled_roles.append(worker.demo_role)
            return use_workers

        eligibility_function = {'func': check_workers_eligibility, 'multiple': True}

        # Assign worker roles is used to determine what the role each worker
        # in the given worker list will play. Setting `id` to None will return
        # the worker to the pool rather than putting them in a given task,
        # which is useful for having tasks with different possible worker
        # counts.
        def assign_worker_roles(workers):
            for worker in workers:
                worker.id = worker.demo_role

        # Define the task function, which will be run with workers that are
        # as the main task.
        global run_conversation

        def run_conversation(mturk_manager, opt, workers):
            # Create the task world
            world = MultiRoleAgentWorld(opt=opt, mturk_agents=workers)
            # run the world to completion
            while not world.episode_done():
                world.parley()

            # shutdown and review the work
            world.shutdown()
            world.review_work()

            # Return the contents for saving
            return world.prep_save_data(workers)

        # Begin the task, allowing mturk_manager to start running the task
        # world on any workers who connect
        mturk_manager.start_task(
            eligibility_function=eligibility_function,
            assign_role_function=assign_worker_roles,
            task_function=run_conversation,
        )
    except BaseException:
        raise
    finally:
        # Any hits that aren't claimed or completed have to be shut down. Must
        # keep the world running until that point.
        mturk_manager.expire_all_unassigned_hits()
        # Shutdown the manager and free all related resources
        mturk_manager.shutdown()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Models and helper classes for style-controlled generation.
"""

import random
from typing import Optional

from parlai.core.message import Message
from parlai.core.opt import Opt


STYLE_SEP_TOKEN = ' STYLE '


class StyleAgentMixin:
    """
    Methods for agents that return style from their histories.
    """

    @classmethod
    def add_cmdline_args(cls, argparser):
        """
        Add command-line arguments specifically for this agent.

        Does not add arguments from its superclass because it's a mixin.
        """
        agent = argparser.add_argument_group('StyleAgentMixin arguments')
        agent.add_argument(
            '--use-style-frac',
            type=float,
            default=1.0,
            help='What fraction of the time to use the style label',
        )
        return agent

    def __init__(self, opt: Opt, shared=None):
        super().__init__(opt, shared)
        self.use_style_frac = opt['use_style_frac']

    def get_temp_history(self, observation: Message) -> Optional[str]:
        """
        Conditionally return a style-token string to temporarily insert into history.
        """
        use_style_rand = random.random()
        if use_style_rand < self.use_style_frac:
            # Use the style
            style = observation.get('personality')
            # This key name is dependent on Image-Chat and will change for other tasks.
            # If obs does not contain 'personality' (i.e. at the end of an epoch during
            # validation), there will be no style
        else:
            style = ''
        if style is not None and style != '':
            return STYLE_SEP_TOKEN + style



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Terminal Chat Runner.

Used to run the terminal chat server.
"""

from parlai.core.params import ParlaiParser
from parlai.chat_service.services.terminal_chat.terminal_manager import TerminalManager
import parlai.chat_service.utils.config as config_utils


SERVICE_NAME = 'terminal'


def setup_args():
    """
    Set up args.

    :return: A parser that takes in command line arguments for chat services (debug, config-path, password), and a port.
    """
    parser = ParlaiParser(False, False)
    parser.add_parlai_data_path()
    parser.add_chatservice_args()
    parser_grp = parser.add_argument_group('Terminal Chat')
    parser_grp.add_argument(
        '--port', default=35496, type=int, help='Port to run the terminal chat server'
    )
    return parser.parse_args()


def run(opt):
    """
    Run TerminalManager.
    """
    opt['service'] = SERVICE_NAME
    manager = TerminalManager(opt)
    try:
        manager.start_task()
    finally:
        manager.shutdown()


if __name__ == '__main__':
    opt = setup_args()
    config_path = opt.get('config_path')
    config = config_utils.parse_configuration_file(config_path)
    opt.update(config['world_opt'])
    opt['config'] = config
    run(opt)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import os
import time
import shlex
from mephisto.core.operator import Operator
from parlai.crowdsourcing.tasks.acute_eval.acute_eval_blueprint import BLUEPRINT_TYPE
from mephisto.utils.scripts import MephistoRunScriptParser

"""
Example script for running ACUTE-EVAL.
The only argument that *must* be modified for this to be run is:
``pairings_filepath``:  Path to pairings file in the format specified in the README.md

The following args are useful to tweak to fit your specific needs;
    - ``annotations_per_pair``: A useful arg if you'd like to evaluate a given conversation pair
                                more than once.
    - ``num_matchup_pairs``:    Essentially, how many pairs of conversations you would like to evaluate
    - ``subtasks_per_unit``:     How many comparisons you'd like a turker to complete in one HIT

"""

TASK_DIRECTORY = os.path.dirname(os.path.realpath(__file__))

parser = MephistoRunScriptParser()
parser.add_argument(
    "-pfp",
    "--pairings-filepath",
    default=f"{TASK_DIRECTORY}/pairings.jsonl",
    help="Path to pairings file",
)
parser.add_argument(
    "-app",
    "--annotations-per-pair",
    default=1,
    help="Annotations per pairing, to ensure worker agreement, default 1",
    type=int,
)
parser.add_argument(
    "-nmp",
    "--num-matchup-pairs",
    default=2,
    help="Number of pairs per model matchup, default 2",
    type=int,
)
parser.add_argument(
    "-spu",
    "--subtasks-per-unit",
    default=5,
    help="Number of conversations to evaluate per task, default 5",
    type=int,
)

architect_type, requester_name, db, args = parser.parse_launch_arguments()

USE_LOCAL = True

task_title = "Which Conversational Partner is Better?"
task_description = "Evaluate quality of conversations through comparison."
hit_keywords = "chat,evaluation,comparison,conversation"

ARG_STRING = (
    f"--blueprint-type {BLUEPRINT_TYPE} "
    f"--architect-type {architect_type} "
    f"--requester-name {requester_name} "
    f'--task-title "\\"{task_title}\\"" '
    f'--task-description "\\"{task_description}\\"" '
    "--task-reward 0.5 "
    f"--task-tags {hit_keywords} "
    f"--maximum-units-per-worker 0 "  # Num of units a worker is allowed to do, 0 is infinite
    f"--allowed-concurrent 1 "  # Workers can only do one task at a time, or onboarding may break
)

extra_args = {
    "pairings_filepath": args['pairings_filepath'],
    "block_on_onboarding_fail": True,
    "block_qualification": f"acute_eval_{int(time.time())}_block",
    # num times to use the same conversation pair
    "annotations_per_pair": args["annotations_per_pair"],
    "random_seed": 42,  # random seed
    "subtasks_per_unit": args[
        "subtasks_per_unit"
    ],  # num comparisons to show within one unit
    "num_matchup_pairs": args[
        "num_matchup_pairs"
    ],  # num pairs of conversations to be compared
    # question phrasing
    "s1_choice": "I would prefer to talk to <Speaker 1>",
    "s2_choice": "I would prefer to talk to <Speaker 2>",
    "eval_question": "Who would you prefer to talk to for a long conversation?",
    "assignment_duration_in_seconds": 600,
}

operator = Operator(db)
operator.parse_and_launch_run_wrapper(shlex.split(ARG_STRING), extra_args=extra_args)
operator.wait_for_runs_then_shutdown(skip_input=True, log_rate=30)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from mephisto.data_model.blueprint import Blueprint
from mephisto.data_model.assignment import InitializationData
from parlai.crowdsourcing.tasks.acute_eval.acute_eval_agent_state import (
    AcuteEvalAgentState,
)
from parlai.crowdsourcing.tasks.acute_eval.acute_eval_runner import AcuteEvalRunner
from parlai.crowdsourcing.tasks.acute_eval.acute_eval_builder import AcuteEvalBuilder
from mephisto.core.registry import register_mephisto_abstraction

import os
import math

from typing import ClassVar, List, Type, Any, Dict, Iterable, TYPE_CHECKING

if TYPE_CHECKING:
    from mephisto.data_model.blueprint import AgentState, TaskRunner, TaskBuilder
    from argparse import _ArgumentGroup as ArgumentGroup

BLUEPRINT_TYPE = "acute_eval"


# WISH AcuteEval's blueprint can probably be extended to compare more than just convos
@register_mephisto_abstraction()
class AcuteEvalBlueprint(Blueprint):
    """
    Blueprint for a task that asks humans to compare conversational outputs.
    """

    AgentStateClass: ClassVar[Type["AgentState"]] = AcuteEvalAgentState
    TaskBuilderClass: ClassVar[Type["TaskBuilder"]] = AcuteEvalBuilder
    TaskRunnerClass: ClassVar[Type["TaskRunner"]] = AcuteEvalRunner
    supported_architects: ClassVar[List[str]] = ["mock"]  # TODO update
    BLUEPRINT_TYPE = BLUEPRINT_TYPE

    @classmethod
    def assert_task_args(cls, opts: Any) -> None:
        """
        Ensure that the data can be properly loaded.
        """
        if opts.get("pairings_filepath") is not None:
            pairings_filepath = os.path.expanduser(opts["pairings_filepath"])
            assert os.path.exists(
                pairings_filepath
            ), f"Provided file {pairings_filepath} doesn't exist"
        elif opts.get("pairings_task_data") is not None:
            assert (
                len(opts.get("pairings_task_data")) > 0
            ), "Length of data dict provided was 0"
        else:
            raise AssertionError(
                "Must provide one of a data csv, json, or a list of tasks"
            )

        if opts.get("block_on_onboarding_fail") is True:
            if opts.get("block_qualification") is None:
                raise AssertionError(
                    "Must provide `block_qualification` to use `block_on_onboarding_fail`"
                )

    @classmethod
    def add_args_to_group(cls, group: "ArgumentGroup") -> None:
        """
        Adds required options for AcuteEvalBlueprints.

        task_source points to the file intending to be deployed for this task
        pairings_filepath has the data to be deployed for this task.
        """
        super(AcuteEvalBlueprint, cls).add_args_to_group(group)

        group.description = """
            AcuteEvalBlueprint: Tasks launched from acute eval blueprints
            require sets of pairings for workers to be able to compare to.

            These pairings can be provided as a csv or by passing a
            pairings_task_data dict into extra_args.
        """
        group.add_argument(
            "--annotations-per-pair",
            dest="annotations_per_pair",
            type=int,
            default=1,
            help="Number of annotations per conversation comparison pair",
        )
        group.add_argument(
            "--pairings-filepath",
            dest="pairings_filepath",
            type=str,
            default=None,
            help="path to the file containing the task dictionaries",
        )
        group.add_argument(
            "--s1-choice",
            dest="s1_choice",
            type=str,
            default="I would prefer to talk to <Speaker 1>",
            help="text next to speaker 1 radio button",
        )
        group.add_argument(
            "--s2-choice",
            dest="s2_choice",
            type=str,
            default="I would prefer to talk to <Speaker 2>",
            help="text next to speaker 2 radio button",
        )
        group.add_argument(
            "--eval-question",
            dest="eval_question",
            type=str,
            default="Who would you prefer to talk to for a long conversation?",
            help='question to present to turker for comparison (e.g. "Which speaker is better?")',
        )
        group.add_argument(
            "--block-on-onboarding-fail",
            dest="block_on_onboarding_fail",
            type=bool,
            default=True,
            help="whether to block on onboarding failure",
        )
        group.add_argument(
            "--subtasks-per-unit",
            dest="subtasks_per_unit",
            type=int,
            default=5,
            help="number of subtasks/comparisons to do per unit",
        )
        group.add_argument(
            "--onboarding-threshold",
            dest="onboarding_threshold",
            type=float,
            default=0.75,
            help="minimum accuracy on onboarding tasks, as a float 0-1.0",
        )
        group.add_argument(
            "--random-seed",
            dest="random_seed",
            type=int,
            default=42,
            help="seed for random",
        )
        group.add_argument(
            "--additional-task-description",
            dest="additional_task_description",
            type=str,
            default='',
            help="Additional text to show on the left pane",
        )
        return

    def get_frontend_args(self) -> Dict[str, Any]:
        """
        Specifies what options within a task_config should be forwarded to the client
        for use by the task's frontend.
        """
        return {
            "task_description": "Placeholder Task Description - Javascript failed to load",
            "frame_height": 650,
            "num_subtasks": self.opts["subtasks_per_unit"],
            "question": self.opts["eval_question"],
            "block_mobile": True,
            "get_task_feedback": False,  # TODO(#95) make option
            "additional_task_description": self.opts['additional_task_description'],
        }

    def get_initialization_data(self) -> Iterable["InitializationData"]:
        """
        Return the InitializationData retrieved from the specified stream.
        """
        # TODO(#99) once we can release HITs over time, configure this to
        # release as many as needed thusfar and top off when
        # onboardings fail
        print(self.opts)
        num_conversations = math.ceil(
            self.opts.get("num_matchup_pairs", 8)
            / max((self.opts["subtasks_per_unit"] - 1), 1)
        )  # release enough hits to finish all annotations requested
        return [
            InitializationData(shared={}, unit_data=[{}])
            for d in range(num_conversations)
        ]



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai import __file__ as parlai_filepath
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.mturk.tasks.acute_eval.run import AcuteEvaluator, add_args as acute_add_args
from parlai.scripts.self_chat import self_chat, setup_args as self_chat_setup_args
from parlai.utils.conversations import Conversations, Conversation
from parlai.utils.strings import normalize_reply
from parlai.utils.testing import capture_output

from parlai.mturk.tasks.acute_eval.analysis import (
    AcuteAnalyzer,
    setup_args as analysis_setup_args,
)
from parlai.mturk.tasks.acute_eval.dump_task_to_acute_format import (
    dump_data as convert_task_data,
    setup_args as convert_task_setup_args,
)
from parlai.mturk.tasks.acute_eval.configs import CONFIG

try:
    from parlai_internal.projects.fast_acute.model_configs import (
        CONFIG as internal_conf,
    )
    from parlai_internal.projects.fast_acute.fast_eval import (
        ACUTE_EVAL_TYPES as internal_types,
    )

    CONFIG.update(internal_conf)
except ImportError:
    # No access to internal
    pass
    internal_types = None

from typing import Dict, Any, List, Tuple, Set
from itertools import combinations

import datetime
import time
import json
import os
import random
import torch
import hashlib

########################
# ACUTE EVAL CONSTANTS #
########################
ACUTE_EVAL_TYPES = {
    'human': {
        'question': 'Which speaker sounds more human?',
        's1_choice': '<Speaker 1> sounds more human',
        's2_choice': '<Speaker 2> sounds more human',
    },
    'engaging': {
        'question': 'Who would you prefer to talk to for a long conversation?',
        's1_choice': 'I would prefer to talk to <Speaker 1>',
        's2_choice': 'I would prefer to talk to <Speaker 2>',
    },
}
if internal_types:
    ACUTE_EVAL_TYPES.update(
        {k: v for k, v in internal_types.items() if k not in ACUTE_EVAL_TYPES}
    )
EXAMPLE_PATH = os.path.join(
    os.path.dirname(parlai_filepath), 'mturk/tasks/acute_eval/example'
)
# Feel free to edit this, but not necessary
SUBTASKS_PER_HIT = 5
MAX_HITS_PER_WORKER = 1
MATCHUPS_PER_PAIR = 160

ACUTE_DEFAULT_ARGS = {
    # onboarding
    'block-on-onboarding-fail': True,
    # pairings
    # general mturk
    'reward': 0.5,
    'max_hits_per_worker': MAX_HITS_PER_WORKER,
    'assignment_duration_in_seconds': 600,
    'auto_approve_delay': 5,
    # acute args
    'annotations_per_pair': 1,
    'seed': 42,
    'subtasks_per_hit': SUBTASKS_PER_HIT,
    # Task Config
    'task_config': {
        'hit_title': 'Which Conversational Partner is Better?',
        'hit_description': 'Evaluate quality of conversations through comparison.',
        'hit_keywords': 'chat,evaluation,comparison,conversation',
    },
    # temp directory for MTURK
    'tmp_dir': '/tmp',
}

#######################
# SELF CHAT CONSTANTS #
#######################
NUM_SELFCHAT_EXAMPLES = 100
SELFCHAT_MAX_TURNS = 6

##############################
# TASK CONVERSTION CONSTANTS #
##############################
NUM_TASK_DATA_EPISODES = 500
SELFCHAT_MAX_TURNS = 6


def setup_args(parser=None) -> ParlaiParser:
    """
    Setup args.
    """
    parser = ParlaiParser(True, False)
    parser.add_argument(
        '--ids',
        type='nonestr',
        help='Comma separated list of CONFIG ids for round robin evaluation (must be at least 2)',
        default=None,
    )
    parser.add_argument(
        '--id-pairs',
        type='nonestr',
        help='Comma separated, colon-delimited list of CONFIG pairs for evaluation, '
        'e.g. model1:model2,model1:model3',
        default=None,
    )
    parser.add_argument(
        '-eval',
        '--acute-eval-type',
        type=str,
        default='engaging',
        choices=list(ACUTE_EVAL_TYPES.keys()),
        help='which evaluation to run for acute',
    )
    parser.add_argument(
        '-mpp',
        '--matchups-per-pair',
        type=int,
        default=MATCHUPS_PER_PAIR,
        help='How many matchups to generate for each pair of ids.',
    )
    parser.add_argument(
        '--live-acute',
        type='bool',
        default=False,
        help='whether this is a LIVE acute run. ',
    )
    parser.add_argument(
        '--onboarding-path',
        type=str,
        default=os.path.join(EXAMPLE_PATH, 'onboarding.jsonl'),
        help='path to onboarding pair',
    )
    parser.set_defaults(selfchat_task=True, task='self_chat')
    return parser


class ParlAIQuickAcute(object):
    """
    Execute Quick ACUTE Runs.
    """

    def __init__(self, opt: Opt):
        self.opt: Opt = opt
        # ids + task
        self._build_id_pairs()
        self.task: str = opt['task']

        # keep track of chat files per model
        self.chat_files: Dict[str, str] = {}

        # question config for running ACUTE
        self.question_config: Dict[str, str] = ACUTE_EVAL_TYPES[opt['acute_eval_type']]

        # root directory for saving everything
        self.root_dir = os.path.join(opt['datapath'], 'acute_eval')
        os.makedirs(self.root_dir, exist_ok=True)

        # path to onboarding file
        self.onboarding_path = opt['onboarding_path']

    #############
    # Utilities #
    #############
    def _build_id_pairs(self):
        """
        Generate self.config_ids and self.combos from self.opt.
        """
        choices = CONFIG.keys()
        combos: Set[Tuple[str, str]] = set()
        ids: Set[str] = set()
        if self.opt['ids'] is None and self.opt['id_pairs'] is None:
            raise RuntimeError(
                'Either --ids or --id-pairs should be set for comparision.'
            )
        if self.opt['id_pairs'] is not None:
            id_pairs = self.opt['id_pairs'].split(',')
            id_pairs = [id_pair.split(':') for id_pair in id_pairs]
            for id_pair in id_pairs:
                combos.add(tuple(sorted((id_pair[0], id_pair[1]))))
                ids |= set(id_pair)
        else:
            ids = set(self.opt['ids'].split(','))
            combos = set(combinations(ids, 2))
        self.config_ids: List[str] = list(ids)
        self.config_ids.sort()
        self.combos: List[Tuple[str, str]] = list(combos)
        self.combos.sort()
        # verify that ids are contained in the config:
        for config_id in self.config_ids:
            if config_id not in choices:
                raise RuntimeError(
                    f'ID {config_id} not specified in the config (`configs.py`).'
                )
        assert len(self.config_ids) > 1, 'Must specify least 2 ids'

    def _print_progress(self, msg: str):
        """
        Format a msg to print to stdout well.

        :param msg:
            message to print
        """
        print(f"\n{'-' * 60}\n {msg} \n {'-' * 60}")

    def _get_selfchat_config(self, config_id: str) -> Dict[str, Any]:
        """
        Return config for selfchat.

        :param config_id:
            config_id string

        :return config:
            dict config for self-chat
        """
        outfile = self._get_selfchat_log_path(config_id)
        config = CONFIG[config_id]
        config.update(
            {
                'task': self.task,
                'outfile': outfile,
                'num_self_chats': NUM_SELFCHAT_EXAMPLES,
                'selfchat_max_turns': SELFCHAT_MAX_TURNS,
                'display_examples': False,
                'log_every_n_secs': -1,
                'indent': -1,
            }
        )
        return config

    def _get_task_conversion_config(self, config_id: str) -> Dict[str, Any]:
        """
        Return config for task conversion to conversation format.

        :param config_id:
            config_id string

        :return config:
            dict config for task conversion script
        """
        outfile = self._get_task_data_path(config_id)
        config = CONFIG[config_id]
        config.update(
            {
                'outfile': outfile,
                'num_episodes': NUM_TASK_DATA_EPISODES,
                'speaker_0_id': f'{config_id}_as_human',
                'speaker_1_id': config_id,
            }
        )
        return config

    def _get_vs_path(self, subdir: str) -> str:
        """
        Return a unique path for the set of comparison combos given a subdirectory.

        We hash the filename as it can grow quite large.
        """
        assert subdir
        os.makedirs(os.path.join(self.root_dir, subdir), exist_ok=True)

        def _combo_name(id1, id2):
            """
            Return joined name for combo of comparisons.
            """
            id1_name = id1
            id2_name = id2
            if 'model' in CONFIG[id1]:
                id1_name += self.task.replace(':', '_')
            if 'model' in CONFIG[id2]:
                id2_name += self.task.replace(':', '_')
            return f'{id1_name}__vs__{id2_name}'

        return os.path.join(
            self.root_dir,
            subdir,
            hashlib.sha1(
                '___and___'.join(
                    [f"{_combo_name(id1, id2)}" for id1, id2 in self.combos]
                ).encode('utf-8')
            ).hexdigest()[:10],
        )

    def _get_log_path(self, config_id: str) -> str:
        """
        Return path to chat logs given config_id.

        :param identifier:
            config_id in CONFIG.
        """
        config = CONFIG[config_id]
        path = ''
        if 'log_path' in config:
            path = config['log_path']
            assert os.path.exists(
                path
            ), f'Path provided in log_path for {config_id} does not exist'
        elif 'task' in config:
            path = self._get_task_data_path(config_id)
        elif 'model' in config:
            path = self._get_selfchat_log_path(config_id)

        assert path, f'Invalid config for {config_id}'

        return path

    def _get_task_data_path(self, config_id: str) -> str:
        """
        Return path to task data as conversations for given task.

        :param config_id:
            config_id string
        """
        task_data_dir = os.path.join(self.root_dir, 'tasks_as_conversations')
        os.makedirs(task_data_dir, exist_ok=True)
        return os.path.join(task_data_dir, f"{config_id}.jsonl")

    def _get_selfchat_log_path(self, config_id: str) -> str:
        """
        Return path to selfchat log for a given model.

        :param config_id:
            config_id string
        """
        self_chat_dir = os.path.join(self.root_dir, 'self_chats')
        os.makedirs(self_chat_dir, exist_ok=True)
        return os.path.join(
            self_chat_dir, f"{config_id}.{self.task.replace(':', '_')}.jsonl"
        )

    def _acutify_convo(
        self, conversation: Conversation, config_id: str
    ) -> Dict[str, List]:
        """
        Format world-logged conversation to be ACUTE format.

        :param conversation:
            dictionary containing the dialogue for a config_id
        :param config_id:
            config_id string

        :return conversation:
            An ACUTE-Readable conversation
        """
        config = CONFIG[config_id]
        is_selfchat = 'model' in config
        acute_conversation: Dict[str, List] = {
            'context': [],
            'dialogue': [],
            'speakers': [],
        }
        for i, ex in enumerate(conversation):
            if ex['id'] == 'context':
                acute_conversation['context'].append(ex)
                continue
            speaker_id = ex['id']
            if is_selfchat:
                speaker_id = config_id if i % 2 == 0 else f'other_{config_id}'
            if speaker_id not in acute_conversation['speakers']:
                acute_conversation['speakers'].append(speaker_id)
            acute_conversation['dialogue'].append(
                {'id': speaker_id, 'text': normalize_reply(ex['text'])}
            )
        return acute_conversation

    def _get_unique_ids(
        self, conversations: Dict[str, Conversations]
    ) -> Dict[str, List[int]]:
        """
        Assign unique IDs for each conversation in conversations.

        This is important for ACUTE-Eval, since we do not want evaluators
        to see the same conversations across comparisons.

        :param conversations:
            Dict mapping config ID to list of conversations

        :return unique_ids:
            dict mapping config id to list of conversation IDs
        """
        id_num = 0
        unique_ids: Dict[str, List[int]] = {}
        for config_id in self.config_ids:
            unique_ids[config_id] = []
            for _convo in conversations[config_id]:
                unique_ids[config_id].append(id_num)
                id_num += 1

        return unique_ids

    def _build_conversation_pairs(
        self, conversations: Dict[str, Conversations]
    ) -> List[Dict[str, Any]]:
        """
        Build a conversation pair to show during ACUTE Eval.

        We build twice as many pairs per matchup as specified
        in the config, to account for issues where sometimes
        we run out of pairs of conversations to evaluate.

        :param conversations:
            A dictionary mapping config_id to dialogues

        :return pairs:
            A list of conversation pairs
        """
        unique_ids = self._get_unique_ids(conversations)
        pairs = []
        pairs_per_id = self.opt['matchups_per_pair'] * 2
        # Write random pairs of conversations
        for id_pair in self.combos:
            for _ in range(pairs_per_id):
                conversation_indices = [
                    random.choice(range(len(conversations[id_]))) for id_ in id_pair
                ]
                pair = []
                pair_ids = []
                for i, c_id in enumerate(conversation_indices):
                    id_ = id_pair[i]
                    pair.append(self._acutify_convo(conversations[id_][c_id], id_))
                    pair_ids.append(unique_ids[id_][c_id])
                pairs.append(
                    {
                        "is_onboarding": False,
                        "speakers_to_eval": id_pair,
                        "dialogue_dicts": pair,
                        "dialogue_ids": pair_ids,
                    }
                )
        return pairs

    def _build_pairings_file(self):
        """
        Build and save pairings to pairings file.
        """
        onboarding_pairs = []
        with open(self.onboarding_path) as f:
            for line in f:
                onboarding_pairs.append(json.loads(line))

        pairings_filepath = self._get_vs_path('pairings_files')

        self._print_progress(f'building pairings file, saving at {pairings_filepath}')
        conversations = {
            config_id: Conversations(self.chat_files[config_id])
            for config_id in self.config_ids
        }
        pairs = self._build_conversation_pairs(conversations)

        with open(pairings_filepath, 'w') as f:
            # Write the onboarding convo
            pairs = onboarding_pairs + pairs
            for pair in pairs:
                f.write(json.dumps(pair) + "\n")

    def _load_pairings_file(self):
        """
        Build the pairings file for ACUTE-Eval.

        If a pairings file already exists, we ask the user whether they would like to
        overwrite the pairings file.
        """
        pairings_filepath = self._get_vs_path('pairings_files')
        # Rebuild pairings file if necessary
        if not os.path.exists(pairings_filepath):
            self._build_pairings_file()
        else:
            modify_time = os.path.getmtime(pairings_filepath)
            self._print_progress(
                f'Pairings already exist {pairings_filepath}. Last modified {time.ctime(modify_time)}'
            )
            answer = ''
            while answer.lower().strip() != 'y' and answer.lower().strip() != 'o':
                answer = input('Enter y to use, o to overwrite:')
                if answer.lower().strip() == 'o':
                    self._build_pairings_file()

        # Load the file
        self._print_progress(f'loading pairings file from {pairings_filepath}')
        self.pairings_filepath = pairings_filepath

    def _run_selfchat(self, config_id: str):
        """
        Run self-chat for model.

        :param config_id:
            id in config
        """
        self._print_progress(f'Running self-chat for {config_id}')
        config = self._get_selfchat_config(config_id)

        with capture_output():
            parser = self_chat_setup_args()
            parser.set_params(**config)
            opt = parser.parse_args(args=[])
        self_chat(opt)

    def _convert_task_to_conversations(self, config_id: str):
        """
        Convert task data to conversations format.

        :param config_id:
            id in config
        """
        self._print_progress(
            f'Converting task data to conversations format for {config_id}'
        )
        config = self._get_task_conversion_config(config_id)

        with capture_output():
            parser = convert_task_setup_args()
            parser.set_params(**config)
            opt = parser.parse_args(args=[])
        convert_task_data(opt)

    ##################
    # Main Functions #
    ##################
    def compile_chat_logs(self):
        """
        Compile Chat Logs.

        Logs are generated depending on what is specified in the config for the model:
        1. If a `model` is provided, run selfchat for model
        2. If a `log_path` is provided, simply load the log path
        3. If a `task` is provided, convert the task to ACUTE format and load that.
        """
        for model in self.config_ids:
            try:
                torch.cuda.empty_cache()
            except Exception:
                pass
            outfile = self._get_log_path(model)

            if not os.path.exists(outfile):
                # 1. Self-chat; 2. Task
                if 'model' in CONFIG[model]:
                    self._run_selfchat(model)
                elif 'task' in CONFIG[model]:
                    self._convert_task_to_conversations(model)
                else:
                    raise RuntimeError(
                        f'Path must exist if log_path specified for {model}'
                    )

                if os.path.exists(outfile):
                    self._print_progress(f'Chats saved to {outfile} for {model}')

            self._print_progress(f'Chats already exist in {outfile}, moving on...')
            self.chat_files[model] = outfile

    def run_acute_eval(self):
        """
        Run ACUTE Eval.
        """
        self._load_pairings_file()

        self.acute_args = acute_add_args()
        self.acute_args.update(ACUTE_DEFAULT_ARGS)
        total_convos = self.opt['matchups_per_pair'] * len(self.combos)
        self.acute_args.update(
            {
                'is_sandbox': not self.opt['live_acute'],
                'pairings_filepath': self.pairings_filepath,
                's1_choice': self.question_config['s1_choice'],
                's2_choice': self.question_config['s2_choice'],
                'question': self.question_config['question'],
                'num_matchup_pairs': total_convos,
                'num_conversations': int(
                    total_convos / (SUBTASKS_PER_HIT - 1)  # subtract 1 for onboarding
                ),
            }
        )
        self.acute_evaluator = AcuteEvaluator(self.acute_args)
        if self.opt['live_acute']:
            self._print_progress('Running ACUTE-EVAL in LIVE Mode')
        else:
            self._print_progress('Running ACUTE-EVAL in SANDBOX Mode')
        self.run_id = self.acute_evaluator.run()

    def analyze_results(self):
        """
        Analyze results of ACUTE Eval run.

        Save results to appropriate filepath.
        """
        self._print_progress(f'Analyzing Results for run id {self.run_id}')
        parser = analysis_setup_args()
        opt = parser.parse_args([])
        today = datetime.date.today().isoformat()
        self.results_path = f"{self._get_vs_path(f'acute_results/{today}/')}"
        opt.update(
            {
                'run_id': self.run_id,
                'outdir': self.results_path,
                'pairings_filepath': self.pairings_filepath,
            }
        )
        opt.update(self.acute_args)

        analyzer = AcuteAnalyzer(opt, self.run_id)
        self.results = analyzer.get_matchup_totals_with_signficance()
        analyzer.save_results()

        self._print_progress(f'ACUTE Results: {self.results}')
        self._print_progress(f'ACUTE results saved to {self.results_path}')


if __name__ == '__main__':
    parser = setup_args()
    runner = ParlAIQuickAcute(parser.parse_args())

    # Compile Chat Logs
    runner.compile_chat_logs()

    # Run ACUTE Eval
    runner.run_acute_eval()

    # Analyze the Results
    runner.analyze_results()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from copy import deepcopy

from parlai.core.worlds import DialogPartnerWorld, validate
from parlai.core.message import Message


class InteractiveWorld(DialogPartnerWorld):
    """
    Simple interactive world involving just two agents talking.

    In more sophisticated worlds the environment could supply information, e.g. in
    tasks/convai2 both agents are given personas, so a world class should be written
    especially for those cases for given tasks.
    """

    def __init__(self, opt, agents, shared=None):
        super().__init__(opt, agents, shared)
        self.init_contexts(shared=shared)
        self.turn_cnt = 0

    def init_contexts(self, shared=None):
        """
        Override to load or instantiate contexts to be used to seed the chat.
        """
        pass

    def get_contexts(self):
        """
        Override to return a pair of contexts with which to seed the episode.

        This function will be called before the first turn of every episode.
        """
        return ['', '']

    def finalize_episode(self):
        print("CHAT DONE ")
        if not self.epoch_done():
            print("\n... preparing new chat... \n")

    def parley(self):
        """
        Agent 0 goes first.

        Alternate between the two agents.
        """
        if self.turn_cnt == 0:
            self.p1, self.p2 = self.get_contexts()

        acts = self.acts
        agents = self.agents
        if self.turn_cnt == 0 and self.p1 != '':
            # add the context on to the first message to agent 0
            context_act = Message(
                {'id': 'context', 'text': self.p1, 'episode_done': False}
            )
            agents[0].observe(validate(context_act))
        try:
            act = deepcopy(agents[0].act())
        except StopIteration:
            self.reset()
            self.finalize_episode()
            self.turn_cnt = 0
            return
        acts[0] = act
        if self.turn_cnt == 0 and self.p2 != '':
            # add the context on to the first message to agent 1
            context_act = Message(
                {'id': 'context', 'text': self.p2, 'episode_done': False}
            )
            agents[1].observe(validate(context_act))
        agents[1].observe(validate(act))
        acts[1] = agents[1].act()
        agents[0].observe(validate(acts[1]))
        self.update_counters()
        self.turn_cnt += 1

        if act['episode_done']:
            self.finalize_episode()
            self.turn_cnt = 0



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.core.build_data import modelzoo_path
from parlai.utils.io import PathManager


def add_cmdline_args(parser):
    # Runtime environment
    agent = parser.add_argument_group('DrQA Arguments')
    agent.add_argument('--no_cuda', type='bool', default=False)
    agent.add_argument('--gpu', type=int, default=-1)
    agent.add_argument('--random_seed', type=int, default=1013)

    # Basics
    agent.add_argument(
        '--embedding_file',
        type=str,
        default=None,
        help='file of space separated embeddings: w e1 ... ed',
    )
    agent.add_argument(
        '--init_model',
        type=str,
        default=None,
        help='Load dict/features/weights/opts from this file',
    )
    agent.add_argument('--log_file', type=str, default=None)

    # Model details
    agent.add_argument('--fix_embeddings', type='bool', default=True)
    agent.add_argument(
        '--tune_partial',
        type=int,
        default=0,
        help='Train the K most frequent word embeddings',
    )
    agent.add_argument(
        '--embedding_dim',
        type=int,
        default=300,
        help=('Default embedding size if ' 'embedding_file is not given'),
    )
    agent.add_argument(
        '--hidden_size', type=int, default=128, help='Hidden size of RNN units'
    )
    agent.add_argument(
        '--doc_layers', type=int, default=3, help='Number of RNN layers for passage'
    )
    agent.add_argument(
        '--question_layers',
        type=int,
        default=3,
        help='Number of RNN layers for question',
    )
    agent.add_argument(
        '--rnn_type',
        type=str,
        default='lstm',
        help='RNN type: lstm (default), gru, or rnn',
    )

    # Optimization details
    agent.add_argument(
        '--valid_metric',
        type=str,
        choices=['accuracy', 'f1'],
        default='f1',
        help='Metric for choosing best valid model',
    )
    agent.add_argument(
        '--max_len', type=int, default=15, help='The max span allowed during decoding'
    )
    agent.add_argument('--rnn_padding', type='bool', default=False)
    agent.add_argument(
        '--display_iter',
        type=int,
        default=10,
        help='Print train error after every' '<display_iter> epoches (default 10)',
    )
    agent.add_argument(
        '--dropout_emb',
        type=float,
        default=0.4,
        help='Dropout rate for word embeddings',
    )
    agent.add_argument(
        '--dropout_rnn', type=float, default=0.4, help='Dropout rate for RNN states'
    )
    agent.add_argument(
        '--dropout_rnn_output',
        type='bool',
        default=True,
        help='Whether to dropout the RNN output',
    )
    agent.add_argument(
        '--optimizer',
        type=str,
        default='adamax',
        help='Optimizer: sgd or adamax (default)',
    )
    agent.add_argument(
        '--learning_rate',
        '-lr',
        type=float,
        default=0.1,
        help='Learning rate for SGD (default 0.1)',
    )
    agent.add_argument(
        '--grad_clipping',
        type=float,
        default=10,
        help='Gradient clipping (default 10.0)',
    )
    agent.add_argument(
        '--weight_decay', type=float, default=0, help='Weight decay (default 0)'
    )
    agent.add_argument('--momentum', type=float, default=0, help='Momentum (default 0)')
    agent.add_argument(
        '--subsample-docs',
        type=int,
        default=0,
        help='When given paragraphs (separated by \n) will take '
        'only a subset of them, including the one with the '
        'answer for training.',
    )

    # Model-specific
    agent.add_argument('--concat_rnn_layers', type='bool', default=True)
    agent.add_argument(
        '--question_merge',
        type=str,
        default='self_attn',
        help='The way of computing question representation',
    )
    agent.add_argument(
        '--use_qemb',
        type='bool',
        default=True,
        help='Whether to use weighted question embeddings',
    )
    agent.add_argument(
        '--use_in_question',
        type='bool',
        default=True,
        help='Whether to use in_question features',
    )
    agent.add_argument(
        '--use_tf', type='bool', default=True, help='Whether to use tf features'
    )
    agent.add_argument(
        '--use_time',
        type=int,
        default=0,
        help='Time features marking how recent word was said',
    )


def set_defaults(opt):
    init_model = None
    # check first for 'init_model' for loading model from file
    if opt.get('init_model') and PathManager.exists(opt['init_model']):
        init_model = opt['init_model']
    # next check for 'model_file', this would override init_model
    if opt.get('model_file') and PathManager.exists(opt['model_file']):
        init_model = opt['model_file']

    if init_model is None:
        # Embeddings options
        opt['embedding_file'] = modelzoo_path(
            opt.get('datapath'), opt['embedding_file']
        )
        if opt.get('embedding_file'):
            if not PathManager.exists(opt['embedding_file']):
                raise IOError('No such file: %s' % opt['embedding_file'])
            with PathManager.open(opt['embedding_file']) as f:
                dim = len(f.readline().strip().split(' ')) - 1
                if dim == 1:
                    # first line was a dud
                    dim = len(f.readline().strip().split(' ')) - 1
            opt['embedding_dim'] = dim
        elif not opt.get('embedding_dim'):
            raise RuntimeError(
                ('Either embedding_file or embedding_dim ' 'needs to be specified.')
            )

        # Make sure tune_partial and fix_embeddings are consistent
        if opt['tune_partial'] > 0 and opt['fix_embeddings']:
            print('Setting fix_embeddings to False as tune_partial > 0.')
            opt['fix_embeddings'] = False

        # Make sure fix_embeddings and embedding_file are consistent
        if opt['fix_embeddings'] and not opt.get('embedding_file'):
            print('Setting fix_embeddings to False as embeddings are random.')
            opt['fix_embeddings'] = False


def override_args(opt, override_opt):
    # Major model args are reset to the values in override_opt.
    # Non-architecture args (like dropout) are kept.
    args = set(
        [
            'embedding_file',
            'embedding_dim',
            'hidden_size',
            'doc_layers',
            'question_layers',
            'rnn_type',
            'optimizer',
            'concat_rnn_layers',
            'question_merge',
            'use_qemb',
            'use_in_question',
            'use_tf',
            'vocab_size',
            'num_features',
            'use_time',
        ]
    )
    for k, v in override_opt.items():
        if k in args:
            opt[k] = v



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Dialogue safety related datasets and teachers.
"""

import parlai.core.build_data as build_data
from parlai.core.message import Message
from parlai.core.teachers import FixedDialogTeacher
from parlai.utils.io import PathManager

from .base_agent import _BaseSafetyTeacher
from .build import build

import copy
import json
import os
import random
import sys as _sys


# Constants
OK_CLASS = '__ok__'
NOT_OK_CLASS = '__notok__'
MULTI_TURN_DATA = 'multi_turn_safety.json'


class StandardTeacher(_BaseSafetyTeacher):
    """
    Data from the standard collection described in the paper `Build it Break it Fix it
    for Dialogue Safety: Robustness from Adversarial Human Attack`
    (<https://arxiv.org/abs/1908.06083>)

    To see data from rounds 1, 2, and 3, try running:
    `parlai display_data -t dialogue_safety:standard --round 3`

    To see data from round 2 only, try running:
    `parlai display_data -t dialogue_safety:standard --round 2
     --round-only True`
    """

    def _load_data_dump(self):
        with PathManager.open(self.data_path, 'rb') as f:
            dump = json.load(f)
        return dump['standard']


class AdversarialTeacher(_BaseSafetyTeacher):
    """
    Data from the adversarial collection described in the paper `Build it Break it Fix
    it for Dialogue Safety: Robustness from Adversarial Human Attack`
    (<https://arxiv.org/abs/1908.06083>)

    To see data from rounds 1, 2, and 3, try running:
    `parlai display_data -t dialogue_safety:adversarial --round 3`

    To see data from round 2 only, try running:
    `parlai display_data -t dialogue_safety:adversarial --round 2
     --round-only True`
    """

    def _load_data_dump(self):
        with PathManager.open(self.data_path, 'rb') as f:
            dump = json.load(f)
        return dump['adversarial']


class MultiturnTeacher(FixedDialogTeacher):
    """
    Data from the multi-turn adversarial collection described in the paper `Build it
    Break it Fix it for Dialogue Safety: Robustness from Adversarial Human Attack`
    (<https://arxiv.org/abs/1908.06083>)

    To see data containing multi-turn conversations, try running
    `parlai display_data -t dialogue_safety:multiturn`.

    Run the above command with the flag `--single-turn True` to only see the
    single turn data.
    """

    @staticmethod
    def add_cmdline_args(parser):
        parser = parser.add_argument_group('Multiturn Safety Teacher Args')
        parser.add_argument(
            '--single-turn',
            type='bool',
            default=False,
            help='only include the single turn data and not the context info',
        )

    def __init__(self, opt, shared=None):
        build(opt['datapath'])  # download the data
        self.opt = opt
        self.data_path = os.path.join(
            opt['datapath'], 'dialogue_safety', MULTI_TURN_DATA
        )

        self.fixed_random = random.Random(42)
        self.single_turn = opt['single_turn']

        if shared and 'data' in shared:
            self.data = shared['data']
        else:
            self._setup_data(opt['datatype'])

        super().__init__(opt, shared)
        self.reset()

    def num_episodes(self):
        return len(self.data)

    def num_examples(self):
        return len(self.data)

    def share(self):
        shared = super().share()
        shared['data'] = self.data
        return shared

    def _setup_data(self, datatype):
        dt = datatype.split(':')[0]
        self.all_data = json.load(open(self.data_path, 'rb'))
        data = self.all_data[dt]
        if self.single_turn:
            # remove
            new_data = []
            for datum in data:
                datum['text'] = datum['text'].split('\n')[-1]
                new_data.append(datum)
            self.data = new_data
        else:
            self.data = data

    def get(self, episode_idx, entry_idx):
        return Message(self.data[episode_idx])


class WikiToxicCommentsTeacher(FixedDialogTeacher):
    """
    Dataset of comments from Wikipedia's Talk page edits. Taken from the Toxic Comments
    Classification Challenge on Kaggle.

    <https://www.kaggle.com/c/jigsaw-toxic-comment-classification-challenge/data>

    We convert this data to a binary classification task.
    """

    @staticmethod
    def add_cmdline_args(parser):
        parser = parser.add_argument_group('Kaggle Toxic Comment Classification Data')
        parser.add_argument(
            '--use-test-set',
            type='bool',
            default=False,
            help='Defaults to 80/10/10 train/test/valid split of train set. '
            'Else, will partition train data into 90/10 train/valid and '
            'use the test set as is.',
        )
        parser.add_argument(
            '--balance-data',
            type='bool',
            default=False,
            help='Balances the data so there are equal numbers of OK and NOT '
            'OK training data',
        )
        return parser

    def __init__(self, opt, shared=None):
        self.fixed_random = random.Random(42)

        self.use_test_set = opt['use_test_set']
        self.balance_data = opt['balance_data']

        self.data_path = os.path.join(
            opt['datapath'], 'dialogue_safety', 'wiki-toxic-comments'
        )

        if shared and 'data' in shared:
            self.data = shared['data']
        else:
            self.build(opt)
            self._setup_data(opt['datatype'])

        self.label_candidates = [NOT_OK_CLASS, OK_CLASS]

        opt = copy.deepcopy(opt)
        super().__init__(opt, shared)
        self.reset()

    def _get_data(self):
        # useful constants
        # all of these colors are bolded
        RESET = '\033[0m'
        RED = '\033[1;91m'
        YELLOW = '\033[1;93m'
        GREEN = '\033[1;92m'
        BLUE = '\033[1;96m'
        CYAN = '\033[1;94m'
        MAGENTA = '\033[1;95m'

        # only use colors if we're outputting to a terminal
        USE_COLORS = _sys.stdout.isatty()
        if not USE_COLORS:
            RESET = RED = YELLOW = GREEN = BLUE = CYAN = MAGENTA = ''

        # generate the rainbow stars
        rainbow = [RED, YELLOW, GREEN, CYAN, BLUE, MAGENTA]
        size = 78 // len(rainbow)
        stars = ''.join([color + '*' * size for color in rainbow])
        stars += RESET

        if not os.path.exists(self.data_path):
            PathManager.mkdirs(self.data_path)
        if not PathManager.exists(os.path.join(self.data_path, 'train.csv')):
            raise RuntimeError(
                f'\n\n{stars}\nThis data must be downloaded from '
                '<https://www.kaggle.com/c/jigsaw-toxic-comment-classification-challenge/data>. '
                '\nIt cannot be automatically downloaded, as one must agree to '
                'the competition rules outlined on the website before '
                'gaining access to the data.\n\n'
                'Once downloaded, please put the data in the following '
                f'directory: \n{self.data_path}\n{stars}'
            )

    def data_to_json(self, pd, file_name):
        response = pd.to_dict('records')
        with PathManager.open(os.path.join(self.data_path, file_name), 'w') as f:
            f.write(json.dumps(response, indent=4))

    def build(self, opt):
        self._get_data()

        try:
            import pandas as pd
        except ImportError:
            raise ImportError('Please install pandas by running `pip install pandas`')

        version = 'v1.0'
        read_path = self.data_path
        if not build_data.built(self.data_path, version):
            print('[building data from : ' + read_path + ']')
            build_data.make_dir(self.data_path)
            # Read in data
            train = pd.read_csv(os.path.join(read_path, 'train.csv'))
            test = pd.read_csv(os.path.join(read_path, 'test.csv'))
            test_labels = pd.read_csv(os.path.join(read_path, 'test_labels.csv'))

            # Labels for test data; value of -1 indicates it was not used for scoring
            test_labels = test_labels[
                (test_labels.toxic != -1)
                & (test_labels.severe_toxic != -1)
                & (test_labels.obscene != -1)
                & (test_labels.threat != -1)
                & (test_labels.insult != -1)
                & (test_labels.identity_hate != -1)
            ]

            # Merge test data with labels
            test = pd.merge(test_labels, test, on='id')

            # Split 10% of train data to be valid
            test['data_type'] = 'test'
            train['data_type'] = 'train'
            valid_set = train.sample(frac=0.1, random_state=42)
            valid_set['data_type'] = 'valid'
            train.update(valid_set)

            # Combine test and train into one data frame
            total_data = pd.concat([test, train], ignore_index=True, sort=False)
            # Rename comment_text to text for the act dict
            total_data.rename(columns={'comment_text': 'text'}, inplace=True)

            # Use the different categories to get binary classification
            total_data['sensitive'] = (
                total_data['severe_toxic']
                + total_data['toxic']
                + total_data['obscene']
                + total_data['threat']
                + total_data['insult']
                + total_data['identity_hate']
            )

            total_data.loc[total_data['sensitive'] < 1, 'is_sensitive'] = 0
            total_data.loc[total_data['sensitive'] >= 1, 'is_sensitive'] = 1

            # Drop unecessary column
            total_data = total_data.drop(columns=['id'])

            self.data_to_json(total_data, 'wiki-toxic-comments-default.json')

            # Partition 80/10/10 according to arXiv:1811.12900 [cs.CL]
            # 90 train 10 test in paper
            # split based on command line flag
            original_train = total_data[
                (total_data['data_type'] == 'valid')
                | (total_data['data_type'] == 'train')
            ].copy()
            l_td = len(original_train)

            original_train.iloc[: int(0.8 * l_td)]['data_type'] = 'train'
            original_train.iloc[int(0.8 * l_td) : int(0.9 * l_td)]['data_type'] = 'test'
            original_train.iloc[int(0.9 * l_td) :]['data_type'] = 'valid'

            self.data_to_json(original_train, 'wiki-toxic-comments-partition.json')

            # Merge the 3 files to get a list of dicts as follows:
            # [
            #     {
            #         'toxic': 0 or 1 ,
            #         'severe_toxic': 0 or 1,
            #         'obscene': 0 or 1,
            #         'threat': 0 or 1,
            #         'insult': 0 or 1,
            #         'identity_hate': 0 or 1,
            #         'text': <comments>,
            #         'data_type': test/validation/train,
            #         'sensitive': 0.0,
            #         'is_sensitive': 0/1
            #     },
            #   ...
            # ]

            build_data.mark_done(self.data_path, version)

    def num_episodes(self):
        return len(self.data)

    def num_examples(self):
        return len(self.data)

    def _balance_data(self, data_list):
        # NOTE: this assumes that there are more OK labels than NOT OK labels
        ok = [x for x in data_list if x['is_sensitive'] == 0]
        notok = [x for x in data_list if x['is_sensitive'] == 1]

        new_not_ok = []
        while len(new_not_ok) < len(ok):
            new_not_ok.append(self.fixed_random.choice(notok))

        new_data = ok + new_not_ok
        self.fixed_random.shuffle(new_data)
        return new_data

    def _setup_data(self, datatype):
        """
        Set up the data based on the correct partition flag specified and partition
        accordingly.
        """
        if not self.use_test_set:
            dp = os.path.join(self.data_path, 'wiki-toxic-comments-partition.json')
        else:
            dp = os.path.join(self.data_path, 'wiki-toxic-comments-default.json')

        print('loading: ' + dp)
        with PathManager.open(dp, 'r') as f:
            self.total_data = json.loads(f.read())
            if 'train' in datatype:
                self.data = [x for x in self.total_data if x['data_type'] == 'train']
            elif 'test' in datatype:
                self.data = [x for x in self.total_data if x['data_type'] == 'test']
            else:
                self.data = [x for x in self.total_data if x['data_type'] == 'valid']

        if self.balance_data and 'train' in datatype:
            self.data = self._balance_data(self.data)

    def get(self, episode_idx, entry_idx):
        sample = self.data[episode_idx]
        sample['label_candidates'] = self.label_candidates
        sample['episode_done'] = True
        sample['labels'] = [self.label_candidates[int(sample['is_sensitive']) - 1]]
        return Message(sample)

    def share(self):
        shared = super().share()
        shared['data'] = self.data
        return shared


class DefaultTeacher(WikiToxicCommentsTeacher):
    pass
